/****************************************************************************\ 
*  Tilt : A Java game for mobile phones                                      *
*  Copyright (C) 2003 Alan Waddington                                        *
*                                                                            *
*  This program is free software; you can redistribute it and/or modify      *
*  it under the terms of the GNU General Public License as published by      *
*  the Free Software Foundation; either version 2 of the License, or         *
*  (at your option) any later version.                                       *
*                                                                            *
*  This program is distributed in the hope that it will be useful,           *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*  GNU General Public License for more details.                              *
*                                                                            *
*  You should have received a copy of the GNU General Public License         *
*  along with this program; if not, write to the Free Software               *
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *  
\****************************************************************************/
/*
 * Game.java
 *
 * Created on 10 May 2003, 07:35
 */

package Tilt;

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import javax.microedition.rms.*;

/**
 *
 * @author  Alan Waddington
 * @version
 */
class Game extends Canvas implements Runnable {
    public Wrapper parent;
    public Board bd;
    private volatile byte direction = 0;              // Last keypress direction
    private boolean terminateGame = false;
    private boolean loadGame = false;
    private boolean saveGame = false;
    public boolean gamePaused = true;      // Game Paused
    public boolean gameOver = false;        // Game Over
    public int score = 0;                   // Game Score
    private Font f;
    private int height;
    private int showScore = 0;
    private boolean clearScreen;
    
    public boolean editMode = false;
    private int colour;
    
    Game(Wrapper p) {
        parent = p;
        bd = new Board(this);
        f = Font.getDefaultFont();
        height = f.getHeight();
    }
    
    public void setLevel(int l) {
        synchronized (bd) {
            bd.level = l;
        }
    }
    
    public int getLevel() {
        synchronized (bd) {
            return bd.level;
        }
    }
    
    public void setEditMode() {
        synchronized(bd) {
            bd.xp = 0;
            bd.yp = 0;
            editMode = true;
            gamePaused = true;
        }
    }
    
    public void setNewGame() {
        synchronized(bd) {
            if (gameOver)
                bd.newLevel(1);
            else
                bd.newLevel(bd.level);
            gameOver = false;
            gamePaused = false;
            score = 0;
            showScore = height;
            clearScreen = true;
            bd.notify();
        }
    }
    
    public void updateScore(int s) {
        score +=s;
        showScore = height;
        if (score<0)
            score = 0;
    }
    
    public void saveGame() {
        synchronized(bd) {
            saveGame = true;
            gamePaused = true;
            bd.notify();
        }
    }
    
    public void loadGame() {
        synchronized(bd) {
            loadGame = true;
            gamePaused = true;
            bd.notify();
        }
    }
    
    public void copyLevel(int level) {
        synchronized (bd) {
            gamePaused = true;
            bd.copyLevel(level);
        }
    }
    
    public void terminateGame() {
        synchronized (bd) {
            terminateGame = true;
            bd.notify();
        }
    }
    
    // Detect EditMode and GameMode Keypresses
    public void keyPressed(int keyCode) {
        int t, icon, icon2;
        int l, x,y;
        int gameAction = getGameAction(keyCode);
        if (editMode) {
            synchronized(bd) {
                l = bd.level;
                x = bd.xp>>3;
                y = bd.yp>>3;
                switch (keyCode) {
                    case Canvas.KEY_NUM0:   // Set Transparent
                        bd.b[l][y][x] = (byte)((5<<2) + colour);
                        break;
                    case Canvas.KEY_NUM1:   // Set Triangle 2
                        bd.b[l][y][x] = (byte)((2<<2) + colour);
                        break;
                    case Canvas.KEY_NUM2:   // Up
                        if (y>0)
                            bd.yp-=8;
                        break;
                    case Canvas.KEY_NUM3:   // Set Triangle 3
                        bd.b[l][y][x] = (byte)((3<<2) + colour);
                        break;
                    case Canvas.KEY_NUM4:   // Left
                        if (x>0)
                            bd.xp-=8;
                        break;
                    case Canvas.KEY_NUM5:   // Set Solid Block
                        bd.b[l][y][x] = (byte)((4<<2) + colour);
                        break;
                    case Canvas.KEY_NUM6:   // Right
                        if (x<bd.X_MAX-1)
                            bd.xp+=8;
                        break;
                    case Canvas.KEY_NUM7:   // Set Triangle 1
                        bd.b[l][y][x] = (byte)((1<<2) + colour);
                        break;
                    case Canvas.KEY_NUM8:   // Down
                        if (y<bd.Y_MAX-1)
                            bd.yp+=8;
                        break;
                    case Canvas.KEY_NUM9:   // Set Triangle 0
                        bd.b[l][y][x] = (byte)((0<<2) + colour);
                        break;
                    case Canvas.KEY_STAR:   // Cycle Colour
                        t = bd.b[l][y][x];
                        if (bd.colours==2)
                            colour = 0;
                        else
                            colour = ((t&3)+1)%4;
                        bd.b[l][y][x] = (byte)((t&0xFC) + colour);
                        break;
                    case Canvas.KEY_POUND:  // Cycle Space, Up, Down
                        t = bd.b[l][y][x];
                        icon = (t>>2);
                        switch (icon) {
                            case 5:
                            case 6:
                                icon2 = icon+1;
                                break;
                            default:
                                icon2 = 5;
                        }
                        bd.b[l][y][x] = (byte)((t&3) + (icon2<<2));
                        break;
                    default:
                        switch (gameAction) {
                            case Canvas.UP:         // Respond to scroll keys
                                if (y>0)
                                    bd.yp-=8;
                                break;
                            case Canvas.DOWN:
                                if (y<bd.Y_MAX-1)
                                    bd.yp+=8;
                                break;
                            case Canvas.LEFT:
                                if (x>0)
                                    bd.xp-=8;
                                break;
                            case Canvas.RIGHT:
                                if (x<bd.X_MAX-1)
                                    bd.xp+=8;
                                break;
                            default:                // All other keys exit
                                editMode = false;
                                parent.reset(true);
                        }
                }
                repaint();
            }
        } else{                         // Game Mode
            switch (keyCode) {
                case Canvas.KEY_NUM0:
                case Canvas.KEY_NUM1:
                case Canvas.KEY_NUM2:
                case Canvas.KEY_NUM3:
                case Canvas.KEY_NUM4:
                case Canvas.KEY_NUM5:
                case Canvas.KEY_NUM6:
                case Canvas.KEY_NUM7:
                case Canvas.KEY_NUM8:
                case Canvas.KEY_NUM9:
                    break;
                case Canvas.KEY_POUND:
                    synchronized (bd) {
                        gamePaused = !gamePaused;
                        if (!gamePaused)
                            bd.notify();
                        showScore = 0;
                        repaint();
                    }
                    break;
                case Canvas.KEY_STAR:
                    synchronized (bd) {
                        gamePaused = true;
                        parent.reset(true);
                    }
                    break;
                default:
            }
            switch (gameAction) {
                case Canvas.UP: // Update current direction
                case Canvas.DOWN:
                case Canvas.LEFT:
                case Canvas.RIGHT:
                    direction = (byte)gameAction;
                    break;
                default:
            }
        }
    }
    
    // Cancel Movement on Key-up
    public void keyReleased(int keyCode) {
        direction = 0;              // Stop moving
    }
    
    // Separate thread that runs the game
    public void run() {
        int FRAME = 100;
        int OVERFRAME = 150;
        long now, delay;
        while (!terminateGame) {
            now = System.currentTimeMillis();
            try {
                synchronized (bd) {
                    if (loadGame) {
                        bd.loadGame();
                        loadGame = false;
                        parent.reset(false);
                        bd.wait();
                    } else if (saveGame) {
                        bd.saveGame();
                        saveGame = false;
                        parent.reset(false);
                        bd.wait();
                    } else if (gameOver) {
                        repaint();
                        bd.wait();
                    } else if (gamePaused) {
                        repaint();
                        bd.wait();
                    } else {
                        bd.move(direction);
                        repaint();
                        delay = FRAME - (System.currentTimeMillis() - now);
                        if (delay>OVERFRAME)
                            bd.wait(delay);
                        else
                            bd.wait(OVERFRAME); // Give the other threads a chance
                    }
                }
            } catch (java.lang.InterruptedException err) { }
        }
    }
    
    // Repaint the screen
    protected void paint(Graphics g) {
        int i, j;
        if (clearScreen) {     // Clear screen in case phone doesn't.
            g.setClip(0, 0, getWidth(), getHeight());
            g.setColor(0x00FFFFFF);
            g.fillRect(0, 0, getWidth(), getHeight());
            g.setClip(0, 0, bd.width, bd.height);
            clearScreen = false;
        }
        bd.paint(g);                    // Paint Board
        if (!editMode) {
            if (gameOver) {
                showScore = 2;
                g.setColor(0x00FFFFFF);
                g.drawString("Game Over", bd.width/2+1,bd.height/2+1,g.TOP|g.HCENTER);
                g.setColor(0x00000000);
                g.drawString("Game Over", bd.width/2,bd.height/2,g.TOP|g.HCENTER);
            } else if (gamePaused) {
                showScore = 2;
                g.setColor(0x00FFFFFF);
                g.drawString("Paused", bd.width/2+1,bd.height/2+1,g.TOP|g.HCENTER);
                g.setColor(0x00000000);
                g.drawString("Paused", bd.width/2,bd.height/2,g.TOP|g.HCENTER);
            }
            if (showScore>-height) {
                showScore--;
                if (showScore>(height>>1))
                    i = 2*showScore-height;
                else if (showScore>0)
                    i = 0;
                else
                    i = -showScore;
                j = height - i;
                g.setColor(0x00FFFFFF);
                g.fillRect(0, 0, bd.width, j+2);
                g.setColor(0x00000000);
                g.drawLine(0,j+2,bd.width-1,j+2);
                g.setColor(0x00000000);
                g.drawString("Score " + score, 4,1-i,g.TOP|g.LEFT);
                g.drawString("Level " + (bd.level), bd.width-4,1-i,g.TOP|g.RIGHT);
            }
        }
    }
}
